/* {=================================================================
 *
 * interval.c
 * int_interval data type for PostBio
 * Author: Luis Carvalho <lexcarvalho at gmail dot com>
 * Please check copyright notice at the bottom of postbio.source
 * $Id: interval.c,v 1.3 2008-01-09 21:27:42 carvalho Exp $
 *
 * ==================================================================} */

#include <postgres.h>
#include <fmgr.h>
#include <access/gist.h>
#include <access/skey.h>

#define INT_INTERVAL_FMT "( %d , %d )"
#define MAXOUTPUT 100

#define MIN(x,y) (x)<(y) ? (x) : (y)
#define MAX(x,y) (x)>(y) ? (x) : (y)

#define OVERLAPS(i1,i2) (i1)->low<=(i2)->high && (i1)->high>=(i2)->low
#define SAME(i1,i2) (i1)->low==(i2)->low && (i1)->high==(i2)->high
#define CONTAINS(i1,i2) (i1)->low<=(i2)->low && (i1)->high>=(i2)->high

#define INT_INTERVAL_CMP_OP(name, op) \
  PG_FUNCTION_INFO_V1(int_ ## name); \
  Datum int_ ## name (PG_FUNCTION_ARGS) { \
    int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0); \
    int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1); \
    if (i1->high == i2->high) PG_RETURN_BOOL(i1->low op i2->low); \
    PG_RETURN_BOOL(i1->high op i2->high); \
  }

#define INT_INTERVAL_TOPO_OP(name, cond) \
  PG_FUNCTION_INFO_V1(int_ ## name); \
  Datum int_ ## name (PG_FUNCTION_ARGS) { \
    int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0); \
    int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1); \
    PG_RETURN_BOOL(cond); \
  }

static char buffer[MAXOUTPUT];

/* integer interval */
typedef struct {
  int32 low;
  int32 high;
} int_interval;

/* {=====================================================================
 *    Prototypes
 * ======================================================================} */

Datum int_in (PG_FUNCTION_ARGS);
Datum int_out (PG_FUNCTION_ARGS);
Datum int_textin (PG_FUNCTION_ARGS);
Datum int_textout (PG_FUNCTION_ARGS);
Datum int_low (PG_FUNCTION_ARGS);
Datum int_high (PG_FUNCTION_ARGS);
Datum integer_to_int (PG_FUNCTION_ARGS);
Datum intpair_to_int (PG_FUNCTION_ARGS);
Datum int_compare (PG_FUNCTION_ARGS);
Datum int_lt (PG_FUNCTION_ARGS);
Datum int_le (PG_FUNCTION_ARGS);
Datum int_eq (PG_FUNCTION_ARGS);
Datum int_ge (PG_FUNCTION_ARGS);
Datum int_gt (PG_FUNCTION_ARGS);
Datum int_left (PG_FUNCTION_ARGS);
Datum int_overleft (PG_FUNCTION_ARGS);
Datum int_overlaps (PG_FUNCTION_ARGS);
Datum int_overright (PG_FUNCTION_ARGS);
Datum int_right (PG_FUNCTION_ARGS);
Datum int_same (PG_FUNCTION_ARGS);
Datum int_contains (PG_FUNCTION_ARGS);
Datum int_contained (PG_FUNCTION_ARGS);
Datum g_int_consistent (PG_FUNCTION_ARGS);
Datum g_int_union (PG_FUNCTION_ARGS);
Datum g_int_compress (PG_FUNCTION_ARGS);
Datum g_int_decompress (PG_FUNCTION_ARGS);
Datum g_int_penalty (PG_FUNCTION_ARGS);
Datum g_int_picksplit (PG_FUNCTION_ARGS);
Datum g_int_same (PG_FUNCTION_ARGS);
Datum int_distance_low (PG_FUNCTION_ARGS);
Datum int_distance_high (PG_FUNCTION_ARGS);
Datum int_distance_center (PG_FUNCTION_ARGS);
Datum int_distance (PG_FUNCTION_ARGS);
Datum int_join (PG_FUNCTION_ARGS);
Datum int_overlap (PG_FUNCTION_ARGS);
Datum int_size (PG_FUNCTION_ARGS);
Datum int_buffer1 (PG_FUNCTION_ARGS);
Datum int_buffer2 (PG_FUNCTION_ARGS);
Datum int_span (PG_FUNCTION_ARGS);


/* {=====================================================================
 *    Functions
 * ======================================================================} */

/* Input/Output */

static int_interval *string2ii (char *str) {
  int32 low, high;
  int_interval *ii;
  if (sscanf(str, INT_INTERVAL_FMT, &low, &high) != 2)
    ereport(ERROR,
        (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
         errmsg("invalid input syntax for int_interval: '%s'", str)));
  ii = (int_interval *) palloc(sizeof(int_interval));
  ii->low = low;
  ii->high = high;
  return ii;
}

static char *ii2string (int_interval *ii) {
  int c = snprintf(buffer, MAXOUTPUT, INT_INTERVAL_FMT, ii->low, ii->high);
  char *str = (char *) palloc(++c);
  memcpy(str, buffer, c);
  return str;
}

PG_FUNCTION_INFO_V1(int_in);
Datum int_in (PG_FUNCTION_ARGS) {
  PG_RETURN_POINTER(string2ii(PG_GETARG_CSTRING(0)));
}

PG_FUNCTION_INFO_V1(int_out);
Datum int_out (PG_FUNCTION_ARGS) {
  PG_RETURN_CSTRING(ii2string((int_interval *) PG_GETARG_POINTER(0)));
}

PG_FUNCTION_INFO_V1(int_textin);
Datum int_textin (PG_FUNCTION_ARGS) {
  text *t = PG_GETARG_TEXT_P(0);
  PG_RETURN_POINTER(string2ii(VARDATA(t)));
}

PG_FUNCTION_INFO_V1(int_textout);
Datum int_textout (PG_FUNCTION_ARGS) {
  int_interval *ii = (int_interval *) PG_GETARG_POINTER(0);
  int c = snprintf(buffer, MAXOUTPUT, INT_INTERVAL_FMT, ii->low, ii->high);
  text *t = (text *) palloc(VARHDRSZ + c);
  VARATT_SIZEP(t) = VARHDRSZ + c;
  memcpy(VARDATA(t), buffer, c);
  PG_RETURN_TEXT_P(t);
}

PG_FUNCTION_INFO_V1(int_low);
Datum int_low (PG_FUNCTION_ARGS) {
  int_interval *ii = (int_interval *) PG_GETARG_POINTER(0);
  int32 low = ii->low;
  PG_RETURN_INT32(low);
}

PG_FUNCTION_INFO_V1(int_high);
Datum int_high (PG_FUNCTION_ARGS) {
  int_interval *ii = (int_interval *) PG_GETARG_POINTER(0);
  int32 high = ii->high;
  PG_RETURN_INT32(high);
}

PG_FUNCTION_INFO_V1(integer_to_int);
Datum integer_to_int (PG_FUNCTION_ARGS) {
  int32 i = PG_GETARG_INT32(0);
  int_interval *ii = (int_interval *) palloc(sizeof(int_interval));
  ii->low = i;
  ii->high = i;
  PG_RETURN_POINTER(ii);
}

PG_FUNCTION_INFO_V1(intpair_to_int);
Datum intpair_to_int (PG_FUNCTION_ARGS) {
  int32 l = PG_GETARG_INT32(0);
  int32 h = PG_GETARG_INT32(1);
  int_interval *ii = (int_interval *) palloc(sizeof(int_interval));
  ii->low = l;
  ii->high = h;
  PG_RETURN_POINTER(ii);
}


/* B-tree (indexing) strategies and support */

PG_FUNCTION_INFO_V1(int_compare); /* support */
Datum int_compare (PG_FUNCTION_ARGS) {
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  if (i1->high == i2->high) PG_RETURN_INT32(i1->low - i2->low);
  PG_RETURN_INT32(i1->high - i2->high);
}

INT_INTERVAL_CMP_OP(lt, <)
INT_INTERVAL_CMP_OP(le, <=)
INT_INTERVAL_CMP_OP(eq, ==)
INT_INTERVAL_CMP_OP(ge, >=)
INT_INTERVAL_CMP_OP(gt, >)


/* R-tree-like GiST (indexing) strategies */

INT_INTERVAL_TOPO_OP(left, i1->high < i2->low)
INT_INTERVAL_TOPO_OP(overleft, i1->high <= i2->high)
INT_INTERVAL_TOPO_OP(overlaps, OVERLAPS(i1, i2))
INT_INTERVAL_TOPO_OP(overright, i1->low >= i2->low)
INT_INTERVAL_TOPO_OP(right, i1->low > i2->high)
INT_INTERVAL_TOPO_OP(same, SAME(i1, i2))
INT_INTERVAL_TOPO_OP(contains, CONTAINS(i1, i2))
INT_INTERVAL_TOPO_OP(contained, CONTAINS(i2, i1))


/* R-tree-like GiST support */

static void int_union_op (int_interval *i1, int_interval *i2,
    int_interval *iu) {
  iu->low = MIN(i1->low, i2->low);
  iu->high = MAX(i1->high, i2->high);
}

static void int_inter_op (int_interval *i1, int_interval *i2,
    int_interval *ii) {
  ii->low = MAX(i1->low, i2->low);
  ii->high = MIN(i1->high, i2->high);
}

static bool g_int_leaf_consistent (int_interval *node,
    int_interval *query, StrategyNumber strategy) {
  bool result;
  switch (strategy) {
    case RTOverlapStrategyNumber:
      result = OVERLAPS(node, query);
      break;
    case RTSameStrategyNumber:
      result = SAME(node, query);
      break;
    case RTContainsStrategyNumber:
    case RTOldContainsStrategyNumber:
      result = CONTAINS(node, query);
      break;
    case RTContainedByStrategyNumber:
    case RTOldContainedByStrategyNumber:
      result = CONTAINS(query, node);
      break;
    default:
      result = FALSE;
  }
  return result;
}

static bool g_int_internal_consistent (int_interval *node,
    int_interval *query, StrategyNumber strategy) {
  bool result;
  switch (strategy) {
    case RTOverlapStrategyNumber:
      result = OVERLAPS(node, query);
      break;
    case RTSameStrategyNumber:
    case RTContainsStrategyNumber:
    case RTOldContainsStrategyNumber:
      result = CONTAINS(node, query);
      break;
    case RTContainedByStrategyNumber:
    case RTOldContainedByStrategyNumber:
      result = OVERLAPS(node, query);
      break;
    default:
      result = FALSE;
  }
  return result;
}

PG_FUNCTION_INFO_V1(g_int_consistent);
Datum g_int_consistent (PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  int_interval *query = (int_interval *) PG_GETARG_POINTER(1);
  StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
  int_interval *node = (int_interval *) DatumGetPointer(entry->key);
  PG_RETURN_BOOL((GIST_LEAF(entry)) /* entry a leaf or internal? */
      ? g_int_leaf_consistent(node, query, strategy)
      : g_int_internal_consistent(node, query, strategy));
}

PG_FUNCTION_INFO_V1(g_int_union);
Datum g_int_union (PG_FUNCTION_ARGS) {
  GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  int_interval *ic = (int_interval *)
    DatumGetPointer(entryvec->vector[0].key);
  int_interval *iu = (int_interval *) palloc(sizeof(int_interval));
  int i;
  iu->low = ic->low;
  iu->high = ic->high;
  for (i = 1; i < entryvec->n; i++)
    int_union_op((int_interval *)
        DatumGetPointer(entryvec->vector[i].key), iu, iu);
  PG_RETURN_POINTER(iu);
}

/* compress and decompress: do nothing */
PG_FUNCTION_INFO_V1(g_int_compress);
Datum g_int_compress (PG_FUNCTION_ARGS) {
  PG_RETURN_DATUM(PG_GETARG_DATUM(0));
}

PG_FUNCTION_INFO_V1(g_int_decompress);
Datum g_int_decompress (PG_FUNCTION_ARGS) {
  PG_RETURN_POINTER(PG_GETARG_POINTER(0));
}

PG_FUNCTION_INFO_V1(g_int_penalty);
Datum g_int_penalty (PG_FUNCTION_ARGS) {
  GISTENTRY *old = (GISTENTRY *) PG_GETARG_POINTER(0);
  GISTENTRY *new = (GISTENTRY *) PG_GETARG_POINTER(1);
  float *result = (float *) PG_GETARG_POINTER(2);
  int_interval *io = (int_interval *) DatumGetPointer(old->key);
  int_interval *in = (int_interval *) DatumGetPointer(new->key);
  int_interval iu;
  int_union_op(io, in, &iu);
  *result = (float) ((iu.high - iu.low) - (io->high - io->low));
  PG_RETURN_FLOAT4(*result); /* size diff */
}

/* Guttman's poly time split algorithm */
PG_FUNCTION_INFO_V1(g_int_picksplit);
Datum g_int_picksplit (PG_FUNCTION_ARGS) {
  GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  GIST_SPLITVEC *splitvec = (GIST_SPLITVEC *) PG_GETARG_POINTER(1);
  OffsetNumber s1, s2, maxoff, *left, *right;
  int_interval *il, *ir, *ic, iul, iur;
  int i, j, lsize, rsize, maxsize, size;
  /* init */
  maxoff = entryvec->n - 2;
  splitvec->spl_left =
    (OffsetNumber *) palloc(entryvec->n * sizeof(OffsetNumber));
  splitvec->spl_nleft = 0;
  splitvec->spl_right =
    (OffsetNumber *) palloc(entryvec->n * sizeof(OffsetNumber));
  splitvec->spl_nright = 0;
  /* select splitting pivots */
  s1 = 1;
  s2 = 2;
  maxsize = 0;
  /* compute max symm diff size over all possible pairs */
  for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i)) {
    il = (int_interval *) DatumGetPointer(entryvec->vector[i].key);
    for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j)) {
      ir = (int_interval *) DatumGetPointer(entryvec->vector[j].key);
      int_union_op(il, ir, &iul);
      size = iul.high - iul.low;
      int_inter_op(il, ir, &iur);
      if (iur.low < iur.high) /* il and ir overlap? */
        size -= iur.high - iur.low;
      if (maxsize < size) {
        s1 = i;
        s2 = j;
        maxsize = size;
      }
    }
  }
  /* split */
  il = (int_interval *) palloc(sizeof(int_interval));
  ic = (int_interval *) DatumGetPointer(entryvec->vector[s1].key);
  il->low = ic->low;
  il->high = ic->high;
  lsize = il->high - il->low;
  ir = (int_interval *) palloc(sizeof(int_interval));
  ic = (int_interval *) DatumGetPointer(entryvec->vector[s2].key);
  ir->low = ic->low;
  ir->high = ic->high;
  rsize = ir->high - ir->low;
  maxoff = OffsetNumberNext(maxoff);
  left = splitvec->spl_left;
  right = splitvec->spl_right;
  for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
    int ulsize, ursize;
    if (i == s1) {
      *left++ = i;
      splitvec->spl_nleft++;
      continue;
    }
    else if (i == s2) {
      *right++ = i;
      splitvec->spl_nright++;
      continue;
    }
    ic = (int_interval *) DatumGetPointer(entryvec->vector[i].key);
    int_union_op(ic, il, &iul);
    ulsize = iul.high - iul.low;
    int_union_op(ic, ir, &iur);
    ursize = iur.high - iur.low;
    /* pick page: update running union and splitvec */
    if (ulsize - lsize < ursize - rsize) { /* left page? */
      il->low = iul.low;
      il->high = iul.high;
      lsize = ulsize;
      *left++ = i;
      splitvec->spl_nleft++;
    }
    else { /* right page */
      ir->low = iur.low;
      ir->high = iur.high;
      rsize = ursize;
      *right++ = i;
      splitvec->spl_nright++;
    }
  }
  *left = *right = FirstOffsetNumber; /* guard */
  splitvec->spl_ldatum = PointerGetDatum(il);
  splitvec->spl_rdatum = PointerGetDatum(ir);
  PG_RETURN_POINTER(splitvec);
}

PG_FUNCTION_INFO_V1(g_int_same);
Datum g_int_same (PG_FUNCTION_ARGS) {
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  PG_RETURN_BOOL(SAME(i1, i2));
}


/* Tools */

PG_FUNCTION_INFO_V1(int_distance_low);
Datum int_distance_low (PG_FUNCTION_ARGS) {
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  PG_RETURN_INT32(abs(i1->low-i2->low));
}

PG_FUNCTION_INFO_V1(int_distance_high);
Datum int_distance_high (PG_FUNCTION_ARGS) {
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  PG_RETURN_INT32(abs(i1->high-i2->high));
}

PG_FUNCTION_INFO_V1(int_distance_center);
Datum int_distance_center (PG_FUNCTION_ARGS) { /* compare function */
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  PG_RETURN_INT32((i1->low+i1->high-i2->low-i2->high)/2);
}

PG_FUNCTION_INFO_V1(int_distance);
Datum int_distance (PG_FUNCTION_ARGS) { /* compare function */
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  PG_RETURN_INT32((abs(i1->low-i2->low)+abs(i1->high-i2->high)
      -(i1->high-i1->low)-(i2->high-i2->low))/2);
}

PG_FUNCTION_INFO_V1(int_join);
Datum int_join (PG_FUNCTION_ARGS) {
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  if (OVERLAPS(i1, i2)) {
    int_interval *iu = (int_interval *) palloc(sizeof(int_interval));
    iu->low = MIN(i1->low, i2->low);
    iu->high = MAX(i1->high, i2->high);
    PG_RETURN_POINTER(iu);
  }
  PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(int_overlap);
Datum int_overlap (PG_FUNCTION_ARGS) {
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  if (OVERLAPS(i1, i2)) {
    int_interval *ii = (int_interval *) palloc(sizeof(int_interval));
    ii->low = MAX(i1->low, i2->low);
    ii->high = MIN(i1->high, i2->high);
    PG_RETURN_POINTER(ii);
  }
  PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(int_size);
Datum int_size (PG_FUNCTION_ARGS) {
  int_interval *ii = (int_interval *) PG_GETARG_POINTER(0);
  PG_RETURN_INT32((ii->low > ii->high) /* valid? */
      ? 0 : ii->high - ii->low + 1);
}

PG_FUNCTION_INFO_V1(int_buffer1);
Datum int_buffer1 (PG_FUNCTION_ARGS) {
  int_interval *ii = (int_interval *) PG_GETARG_POINTER(0);
  int32 offset = PG_GETARG_INT32(1);
  int_interval *ib;
  if (offset < 0 || offset > ii->low) PG_RETURN_NULL();
  ib = (int_interval *) palloc(sizeof(int_interval));
  ib->low = ii->low - offset;
  ib->high = ii->high + offset;
  PG_RETURN_POINTER(ib);
}

PG_FUNCTION_INFO_V1(int_buffer2);
Datum int_buffer2 (PG_FUNCTION_ARGS) {
  int_interval *ii = (int_interval *) PG_GETARG_POINTER(0);
  int32 l_offset = PG_GETARG_INT32(1);
  int32 h_offset = PG_GETARG_INT32(2);
  int_interval *ib;
  if (l_offset < 0 || l_offset > ii->low || h_offset < 0)
    PG_RETURN_NULL();
  ib = (int_interval *) palloc(sizeof(int_interval));
  ib->low = ii->low - l_offset;
  ib->high = ii->high + h_offset;
  PG_RETURN_POINTER(ib);
}

PG_FUNCTION_INFO_V1(int_span);
Datum int_span (PG_FUNCTION_ARGS) {
  int_interval *i1 = (int_interval *) PG_GETARG_POINTER(0);
  int_interval *i2 = (int_interval *) PG_GETARG_POINTER(1);
  int_interval *iu = (int_interval *) palloc(sizeof(int_interval));
  iu->low = MIN(i1->low, i2->low);
  iu->high = MAX(i1->high, i2->high);
  PG_RETURN_POINTER(iu);
}

