/* {=================================================================
 *
 * stree.c
 * Suffix tree for MUMs in PostBio
 * Author: Luis Carvalho <lexcarvalho at gmail dot com>
 * Please check copyright notice at the bottom of postbio.source
 * $Id: stree.c,v 1.2 2008-01-12 04:52:34 carvalho Exp $
 *
 * ==================================================================} */

#include <postgres.h>
#include <fmgr.h>
#include <funcapi.h>
#include <miscadmin.h>
#include <access/heapam.h>
#include <utils/builtins.h>

#include <stree/types.h>
#include <stree/symboldef.h>
#include <stree/streedef.h>
#include <stree/protodef.h>

/* All these definitions are either copied or adapted from Stefan Kurtz's
 * libstree for a more memory-efficient implementation of a Set-Returning
 * Function (SRF) like mum below; for license details please refer to
 * stree/LICENSE * */

#define LARGESTCHARINDEX UCHAR_MAX

/* stree */
Sint constructstree (Suffixtree *stree, SYMBOL *text, Uint textlen);
void freestree (Suffixtree *stree);

/* maxmat */
typedef Sint (*Processmatchfunction) (void *, Uint, Uint, Uint, Uint);
Sint findmumcandidates (Suffixtree *stree, Uint minmatchlength,
    Processmatchfunction processmatch, void *processinfo,
    Uchar *query, Uint querylen, Uint seqnum);
Sint findmaxmatches (Suffixtree *stree, Uint minmatchlength,
    Processmatchfunction processmatch, void *processinfo,
    Uchar *query, Uint querylen, Uint seqnum);

/* From mm3rc/findmaxmat.c */
typedef struct {
  Uint querycommondepth;
  BOOL onmaxpath;
} Nodeinfo;

DECLAREARRAYSTRUCT(Nodeinfo);

/* This is Maxmatchinfo from mm3rc/findmaxmat.c plus: mumcandtab, mumcand,
 * loc, left, right, uniqueinref, hasreturned, and result fields */
typedef struct MUM_Info {
  /* findmaxmatches */
  Suffixtree *stree;
  ArrayNodeinfo commondepthstack;
  ArrayPathinfo matchpath;
  Location maxloc;
  Uchar *query, *querysuffix;
  Uint querylen, minmatchlength, depthofpreviousmaxloc;
  ArrayMUMcandidate mumcandtab;
  MUMcandidate *mumcand;
  /* findmumcand */
  Location loc;
  Uchar *left, *right;
  /* global */
  bool uniqueinref, hasreturned;
  Datum result;
} MUM_Info;

#define islocationMUMcand(M) \
  ((M)->loc.remain > 0 && (M)->loc.nextnode.toleaf \
   && ((M)->querysuffix == (M)->query \
     || (M)->loc.locstring.start == 0 \
     || *((M)->querysuffix - 1) \
       != (M)->stree->text[(M)->loc.locstring.start - 1]))

/* end of libstree */

/* Main datatype */
typedef struct Stree {
  int32 size; /* varlena */
  Suffixtree tree;
} Stree;

/* Used when returning match tuples */
static bool nulls[] = {0, 0, 0};

typedef struct MUM_Count {
  int nmatches;
  bool anymatch;
} MUM_Count;


/* {=====================================================================
 *    Prototypes
 * ======================================================================} */

Datum stree_in (PG_FUNCTION_ARGS);
Datum stree_out (PG_FUNCTION_ARGS);
Datum stree_build (PG_FUNCTION_ARGS);
Datum stree_text (PG_FUNCTION_ARGS);
Datum stree_subtext (PG_FUNCTION_ARGS);
Datum maxmatch (PG_FUNCTION_ARGS);
Datum maxmatchcount (PG_FUNCTION_ARGS);
Datum maxmatch0 (PG_FUNCTION_ARGS);


/* {=====================================================================
 *    Functions
 * ======================================================================} */

static Uint isseqvalid (Uchar *seq, Uint l) {
  Uchar *s;
  for (s = seq; s < seq + l; s++) {
    switch (*s) {
      case 'a': case 'c': case 'g': case 't': case 's':
      case 'w': case 'r': case 'y': case 'm': case 'k':
      case 'b': case 'd': case 'h': case 'v': case 'n':
        break;
      default: /* illegal char */
        return 0;
    }
  }
  return l;
}

/* update pointers */
static void initstree (Stree *S) {
  Uint l = S->tree.textlen;
  S->tree.text = (SYMBOL *) (S + 1);
  S->tree.sentinel = S->tree.text + l;
  S->tree.leaftab = (Uint *) S->tree.sentinel;
  S->tree.rootchildren = S->tree.leaftab + l + 2;
  S->tree.branchtab = S->tree.rootchildren + LARGESTCHARINDEX + 1;
  S->tree.nextfreebranch = S->tree.branchtab + S->tree.nextfreebranchnum;
  S->tree.chainstart = (S->tree.chainstart) ? S->tree.nextfreebranch : NULL;
}

/* copy Suffixtree with minimal pointers for MUM search */
static Datum buildstree (SYMBOL *text, Uint l) {
  Suffixtree stree;
  Stree *S;
  Uint i;
  int32 size;
  if (!isseqvalid(text, l))
    elog(ERROR, "invalid symbol or empty sequence");
  if (l > getmaxtextlenstree())
    elog(ERROR, "sequence too large for suffix tree construction");
  if (constructprogressstree(&stree, text, l, NULL, NULL, NULL))
    elog(ERROR, "error building suffix tree");
  size = sizeof(Stree)
    + l /* text */
    + (l + 2 /* leaftab */
    + LARGESTCHARINDEX + 1 /* rootchildren */
    + stree.currentbranchtabsize) * sizeof(Uint); /* branchtab */
  S = (Stree *) palloc(size);
  S->size = size;
  S->tree.textlen = l;
  initstree(S); /* fix references */
  /* copy from stree */
  memcpy(S->tree.text, text, l);
  S->tree.currentdepth = stree.currentdepth;
  S->tree.nextfreeleafnum = stree.nextfreeleafnum;
  S->tree.nextfreebranchnum = stree.nextfreebranchnum;
  S->tree.currentbranchtabsize = stree.currentbranchtabsize;
  /* leaftab */
  for (i = 0; i < l + 2; i++)
    S->tree.leaftab[i] = stree.leaftab[i];
  /* rootchildren */
  for (i = 0; i < LARGESTCHARINDEX + 1; i++)
    S->tree.rootchildren[i] = stree.rootchildren[i];
  /* branchtab */
  for (i = 0; i < S->tree.currentbranchtabsize; i++)
    S->tree.branchtab[i] = stree.branchtab[i];
  /* nextfreebranch and chainstart */
  S->tree.nextfreebranch = S->tree.branchtab + S->tree.nextfreebranchnum;
  S->tree.chainstart = (stree.chainstart) ? S->tree.nextfreebranch : NULL;
  freestree(&stree);
  PG_RETURN_POINTER(S);
}

PG_FUNCTION_INFO_V1(stree_in);
Datum stree_in (PG_FUNCTION_ARGS) {
  SYMBOL *text = (SYMBOL *) PG_GETARG_CSTRING(0);
  Uint l = (Uint) strlen((const char *) text);
  return buildstree(text, l);
}

PG_FUNCTION_INFO_V1(stree_out);
Datum stree_out (PG_FUNCTION_ARGS) {
  Stree *S = (Stree *) PG_GETARG_VARLENA_P(0);
  char *result = (char *) palloc(S->tree.textlen + 1);
  memcpy(result, S + 1, S->tree.textlen);
  result[S->tree.textlen] = '\0';
  PG_RETURN_CSTRING(result);
}

PG_FUNCTION_INFO_V1(stree_build);
Datum stree_build (PG_FUNCTION_ARGS) {
  text *t = PG_GETARG_TEXT_P(0);
  SYMBOL *text = (SYMBOL *) VARDATA(t);
  Uint l = VARSIZE(t) - VARHDRSZ;
  return buildstree(text, l);
}

PG_FUNCTION_INFO_V1(stree_text);
Datum stree_text (PG_FUNCTION_ARGS) {
  Stree *S = (Stree *) PG_GETARG_VARLENA_P(0);
  text *result = (text *) palloc(VARHDRSZ + S->tree.textlen);
  VARATT_SIZEP(result) = VARHDRSZ + S->tree.textlen;
  memcpy(VARDATA(result), S + 1, S->tree.textlen);
  PG_RETURN_TEXT_P(result);
}

PG_FUNCTION_INFO_V1(stree_subtext);
Datum stree_subtext (PG_FUNCTION_ARGS) {
  Stree *S = (Stree *) PG_GETARG_VARLENA_P(0);
  int start = PG_GETARG_INT32(1);
  int len = PG_GETARG_INT32(2);
  text *result;
  if (start <= 0 || len <= 0)
    elog(ERROR, "non-positive start or length");
  start--; /* zero-based */
  if (start + len > S->tree.textlen)
    len = S->tree.textlen - start;
  result = (text *) palloc(VARHDRSZ + len);
  VARATT_SIZEP(result) = VARHDRSZ + len;
  memcpy(VARDATA(result), ((char *)(S + 1)) + start, len);
  PG_RETURN_TEXT_P(result);
}


/* ======= findmaxmatches ======= */

/* Adapted from mm3rc/findmaxmat.c */
static Sint processleaf (Uint leafindex, Bref lcpnode, void *info) {
  MUM_Info *muminfo = (MUM_Info *) info;
  (void) lcpnode; /* unused */
  if (leafindex == 0
      || muminfo->query == muminfo->querysuffix
      || muminfo->stree->text[leafindex - 1]
        != *(muminfo->querysuffix - 1)) {
    Uint lcplength;
    MUMcandidate *mumcand;
    if (muminfo->commondepthstack.nextfreeNodeinfo == 0)
      lcplength = muminfo->maxloc.locstring.length;
    else {
      Nodeinfo *father = muminfo->commondepthstack.spaceNodeinfo
        + muminfo->commondepthstack.nextfreeNodeinfo - 1;
      if (father->onmaxpath &&
          LEAFADDR2NUM(muminfo->stree,
                      muminfo->maxloc.nextnode.address) == leafindex)
        lcplength = muminfo->maxloc.locstring.length;
      else
        lcplength = father->querycommondepth;
    }
    /* process MUM candidate */
    GETNEXTFREEINARRAY(mumcand, &muminfo->mumcandtab, MUMcandidate, 1024);
    mumcand->mumlength = lcplength; /* match length */
    mumcand->dbstart = leafindex; /* subject start */
    mumcand->querystart = (Uint) (muminfo->querysuffix
        - muminfo->query); /* query start */
  }
  return 0;
}

/* Copied from mm3rc/findmaxmat.c */
static void inheritfrompath (ArrayPathinfo *matchpath, Location *maxloc,
    Nodeinfo *stacktop, Bref nodeptr, Uint accessindex, Uint inheritdepth) {
  if (accessindex > matchpath->nextfreePathinfo) {
    stacktop->onmaxpath = False;
    stacktop->querycommondepth = inheritdepth;
  }
  else {
    if (accessindex == matchpath->nextfreePathinfo) {
      if (maxloc->remain == 0) {
        if (maxloc->nextnode.address == nodeptr) {
          stacktop->onmaxpath = True;
          stacktop->querycommondepth = maxloc->locstring.length;
        }
        else {
          stacktop->onmaxpath = False;
          stacktop->querycommondepth = inheritdepth;
        }
      }
      else {
        stacktop->onmaxpath = False;
        if (maxloc->nextnode.address == nodeptr)
          stacktop->querycommondepth = maxloc->locstring.length;
        else
          stacktop->querycommondepth = inheritdepth;
      }
    }
    else {
      if (matchpath->spacePathinfo[accessindex].ref == nodeptr) {
        stacktop->onmaxpath = True;
        stacktop->querycommondepth =
          matchpath->spacePathinfo[accessindex].depth;
      }
      else {
        stacktop->onmaxpath = False;
        stacktop->querycommondepth = inheritdepth;
      }
    }
  }
}

/* Copied from mm3rc/findmaxmat.c */
static BOOL processbranch1 (Bref nodeptr, void *info) {
  MUM_Info *muminfo = (MUM_Info *) info;
  Nodeinfo *stacktop, *father;
  GETNEXTFREEINARRAY(stacktop, &muminfo->commondepthstack, Nodeinfo, 32);
  if (stacktop == muminfo->commondepthstack.spaceNodeinfo)
    inheritfrompath(&muminfo->matchpath, &muminfo->maxloc, stacktop,
        nodeptr, 0, muminfo->minmatchlength);
  else {
    father = stacktop-1;
    if (father->onmaxpath)
      inheritfrompath(&muminfo->matchpath, &muminfo->maxloc, stacktop,
          nodeptr, muminfo->commondepthstack.nextfreeNodeinfo - 1,
          father->querycommondepth);
    else {
      stacktop->onmaxpath = False;
      stacktop->querycommondepth = father->querycommondepth;
    }
  }
  return True;
}

/* Copied from mm3rc/findmaxmat.c */
static Sint processbranch2 (Bref nodeptr, void *info) {
  MUM_Info *muminfo = (MUM_Info *) info;
  (void) nodeptr; /* unused */
  muminfo->commondepthstack.nextfreeNodeinfo--;
  return 0;
}

/* Adapted from mm3rc/findmaxmat.c */
static Sint enumeratemaxmatches (MUM_Info *muminfo, Location *ploc) {
  Uint rescanprefixlength;
  muminfo->matchpath.nextfreePathinfo = 0;
  if (muminfo->depthofpreviousmaxloc > UintConst(1))
    rescanprefixlength = muminfo->depthofpreviousmaxloc - 1;
  else
    rescanprefixlength = 0;
  findprefixpathstree(muminfo->stree, &muminfo->matchpath, &muminfo->maxloc,
      ploc, muminfo->querysuffix + muminfo->minmatchlength,
      muminfo->query + muminfo->querylen - 1, rescanprefixlength);
  muminfo->depthofpreviousmaxloc = muminfo->maxloc.locstring.length;
  muminfo->commondepthstack.nextfreeNodeinfo = 0;
  if (ploc->nextnode.toleaf)
    processleaf(LEAFADDR2NUM(muminfo->stree, ploc->nextnode.address),
        NULL, (void *) muminfo);
  else {
    processbranch1(ploc->nextnode.address, (void *) muminfo);
    depthfirststree(muminfo->stree, &ploc->nextnode, processleaf,
        processbranch1, processbranch2, NULL, NULL, (void *) muminfo);
  }
  return 0;
}


static int processmaxmatches (MUM_Info *muminfo, TupleDesc tupdesc) {
  if (!muminfo->hasreturned) { /* grab some matches? */
    enumeratemaxmatches(muminfo, &muminfo->loc);
    muminfo->mumcand = muminfo->mumcandtab.spaceMUMcandidate;
  }
  if (muminfo->mumcand < muminfo->mumcandtab.spaceMUMcandidate
      + muminfo->mumcandtab.nextfreeMUMcandidate) { /* any match? */
    Datum values[3];
    values[0] = muminfo->mumcand->mumlength; /* match length */
    values[1] = muminfo->mumcand->dbstart + 1; /* subject start */
    values[2] = muminfo->mumcand->querystart + 1; /* query start */
    muminfo->result = HeapTupleGetDatum(heap_form_tuple(tupdesc,
          values, nulls));
    muminfo->mumcand++;
    muminfo->hasreturned = 1;
  }
  else {
    muminfo->mumcandtab.nextfreeMUMcandidate = 0; /* reset mumcandtab */
    muminfo->hasreturned = 0; /* reset */
  }
  return muminfo->hasreturned;
}

/* Adapted from mm3rc/findmaxmat.c */
static void findmaxmatchessrf (MUM_Info *muminfo, TupleDesc tupdesc) {
  while (muminfo->right < muminfo->query + muminfo->querylen - 1) {
    if (muminfo->loc.locstring.length >= muminfo->minmatchlength) {
      if (processmaxmatches(muminfo, tupdesc)) return;
    }
    if (ROOTLOCATION(&muminfo->loc))
      scanprefixfromnodestree(muminfo->stree, &muminfo->loc,
          ROOT(muminfo->stree), muminfo->querysuffix + 1,
          muminfo->right + 1, 0);
    else {
      linklocstree(muminfo->stree, &muminfo->loc, &muminfo->loc);
      scanprefixstree(muminfo->stree, &muminfo->loc, &muminfo->loc,
          muminfo->querysuffix + muminfo->loc.locstring.length + 1,
          muminfo->right + 1, 0);
    }
    muminfo->querysuffix++;
    muminfo->right++;
  }
  while (!ROOTLOCATION(&muminfo->loc)
      && muminfo->loc.locstring.length >= muminfo->minmatchlength) {
    if (processmaxmatches(muminfo, tupdesc)) return;
    linklocstree(muminfo->stree, &muminfo->loc, &muminfo->loc);
    muminfo->querysuffix++;
  }
  muminfo->result = (Datum) 0;
}


/* ======= findmumcand ======= */

static int processmumcand (MUM_Info *muminfo, TupleDesc tupdesc) {
  if (muminfo->hasreturned) /* already processed these matches? */
    muminfo->hasreturned = 0; /* reset */
  else {
    Datum values[3];
    values[0] = muminfo->loc.locstring.length; /* match length */
    values[1] = muminfo->loc.locstring.start + 1; /* subject start */
    values[2] = muminfo->querysuffix - muminfo->query + 1; /* query start */
    muminfo->result = HeapTupleGetDatum(heap_form_tuple(tupdesc, values, nulls));
    muminfo->hasreturned = 1;
  }
  return muminfo->hasreturned;
}

/* Adapted from mm3rc/findmumcand.c */
static void findmumcandsrf (MUM_Info *muminfo, TupleDesc tupdesc) {
  while (muminfo->left != NULL) {
    if (muminfo->loc.locstring.length >= muminfo->minmatchlength
        && islocationMUMcand(muminfo)) { /* mum candidate? */
      if (processmumcand(muminfo, tupdesc)) return; /* process it */
    }
    if (ROOTLOCATION(&muminfo->loc))
      muminfo->left = scanprefixfromnodestree(muminfo->stree, &muminfo->loc,
          ROOT(muminfo->stree), muminfo->left + 1, muminfo->right, 0);
    else {
      linklocstree(muminfo->stree, &muminfo->loc, &muminfo->loc);
      muminfo->left = scanprefixstree(muminfo->stree, &muminfo->loc,
          &muminfo->loc, muminfo->left, muminfo->right, 0);
    }
    muminfo->querysuffix++;
  }
  while (!ROOTLOCATION(&muminfo->loc)
      && muminfo->loc.locstring.length >= muminfo->minmatchlength) {
    if (islocationMUMcand(muminfo)) { /* mum candidate? */
      if (processmumcand(muminfo, tupdesc)) return; /* process it */
    }
    linklocstree(muminfo->stree, &muminfo->loc, &muminfo->loc);
    muminfo->querysuffix++;
  }
  muminfo->result = (Datum) 0;
}

/* Here we adapt the original implementation of findmaxmatches and
 * findmumcandidates to store batches of MUMs in a MUMcandidate array; this is
 * more memory efficient since we don't need to store _all_ results in the
 * array and then duplicate them when materializing the query result */

PG_FUNCTION_INFO_V1(maxmatch);
Datum maxmatch (PG_FUNCTION_ARGS) {
  FuncCallContext *funcctx;
  MUM_Info *muminfo;
  if (SRF_IS_FIRSTCALL()) {
    MemoryContext oldcontext;
    Stree *S = (Stree *) PG_GETARG_VARLENA_P(0);
    text *q = PG_GETARG_TEXT_P(1);
    Uchar *query = (Uchar *) VARDATA(q);
    Uint querylen = VARSIZE(q) - VARHDRSZ;
    bool uniqueinref = PG_GETARG_BOOL(2);
    Uint minmatchlength = (Uint) PG_GETARG_UINT32(3);
    if (!isseqvalid(query, querylen))
      elog(ERROR, "invalid symbol or empty sequence");
    if (minmatchlength == 0) minmatchlength = querylen;
    /* setup funcctx */
    funcctx = SRF_FIRSTCALL_INIT();
    if (querylen < minmatchlength) /* early return? */
      SRF_RETURN_DONE(funcctx);
    oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
    if (get_call_result_type(fcinfo, NULL, &funcctx->tuple_desc)
        != TYPEFUNC_COMPOSITE)
      elog(ERROR, "function returning record called in context "
          "that cannot accept type record");
    funcctx->user_fctx = palloc(sizeof(MUM_Info) + S->size);
    muminfo = (MUM_Info *) funcctx->user_fctx;
    memcpy(muminfo + 1, S, S->size); /* to avoid wrong refs */
    S = (Stree *) (muminfo + 1);
    initstree(S);
    muminfo->stree = &S->tree;
    MemoryContextSwitchTo(oldcontext);
    /* setup muminfo */
    muminfo->querysuffix = muminfo->query = query;
    muminfo->querylen = querylen;
    muminfo->minmatchlength = minmatchlength;
    muminfo->uniqueinref = uniqueinref;
    if (uniqueinref)
      muminfo->right = muminfo->query + muminfo->querylen - 1;
    else {
      muminfo->right = muminfo->query + muminfo->minmatchlength - 1;
      INITARRAY(&muminfo->commondepthstack, Nodeinfo);
      INITARRAY(&muminfo->matchpath, Pathinfo);
      INITARRAY(&muminfo->mumcandtab, MUMcandidate);
      muminfo->mumcand = NULL;
    }
    muminfo->left = scanprefixfromnodestree(muminfo->stree, &muminfo->loc,
        ROOT(muminfo->stree), muminfo->query, muminfo->right, 0);
    muminfo->depthofpreviousmaxloc = muminfo->loc.locstring.length;
    muminfo->result = (Datum) 0;
    muminfo->hasreturned = 0;
  }
  funcctx = SRF_PERCALL_SETUP();
  muminfo = (MUM_Info *) funcctx->user_fctx;
  /* find MUMs */
  if (muminfo->uniqueinref)
    findmumcandsrf(muminfo, funcctx->tuple_desc);
  else
    findmaxmatchessrf(muminfo, funcctx->tuple_desc);
  if (muminfo->result)
    SRF_RETURN_NEXT(funcctx, muminfo->result);
  else {
    if (!muminfo->uniqueinref) {
      FREEARRAY(&muminfo->commondepthstack, Nodeinfo);
      FREEARRAY(&muminfo->matchpath, Pathinfo);
      FREEARRAY(&muminfo->mumcandtab, MUMcandidate);
    }
    SRF_RETURN_DONE(funcctx);
  }
}


/* ======= maxmatch0 ======= */

/* This is a previous implementation of mum, and we keep it here for
 * historical/pedagogical reasons. It is the most memory efficient
 * implementation since MUMs are stored directly in the query result. However,
 * this is a _table function_, that is, we can only use it in queries like
 * "select * from maxmatch0(...)". */

static Sint processmum (void *info, Uint matchlength,
    Uint subjectstart, Uint seqnum, Uint querystart) {
  Datum values[3];
  ReturnSetInfo *rsi = (ReturnSetInfo *) info;
  MemoryContext mcxt;
  values[0] = matchlength;
  values[1] = subjectstart + 1;
  (void) seqnum; /* unused */
  values[2] = querystart + 1;
  mcxt = MemoryContextSwitchTo(rsi->econtext->ecxt_per_query_memory);
  tuplestore_puttuple(rsi->setResult,
      heap_form_tuple(rsi->setDesc, values, nulls));
  MemoryContextSwitchTo(mcxt);
  return 0;
}

PG_FUNCTION_INFO_V1(maxmatch0);
Datum maxmatch0 (PG_FUNCTION_ARGS) {
  MemoryContext mcxt;
  ReturnSetInfo *rsi = (ReturnSetInfo *) fcinfo->resultinfo;
  Stree *S = (Stree *) PG_GETARG_VARLENA_P(0);
  text *q = PG_GETARG_TEXT_P(1);
  Uchar *query = (Uchar *) VARDATA(q);
  Uint querylen = VARSIZE(q) - VARHDRSZ;
  bool uniqueinref = PG_GETARG_BOOL(2);
  Uint minmatchlength = (Uint) PG_GETARG_UINT32(3);
  if (!isseqvalid(query, querylen))
    elog(ERROR, "invalid symbol or empty sequence");
  if (minmatchlength == 0) minmatchlength = querylen;
  if (querylen < minmatchlength) /* early return? */
    PG_RETURN_NULL();
  /* prepare tuple set and tree */
  if (!rsi || !IsA(rsi, ReturnSetInfo)
      || (rsi->allowedModes & SFRM_Materialize) == 0
      || (rsi->expectedDesc == NULL))
    ereport(ERROR,
        (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
         errmsg("set-valued function called in context "
           "that cannot accept a set")));
  rsi->returnMode = SFRM_Materialize;
  mcxt = MemoryContextSwitchTo(rsi->econtext->ecxt_per_query_memory);
  rsi->setResult = tuplestore_begin_heap(true, false, work_mem);
  rsi->setDesc = CreateTupleDescCopy(rsi->expectedDesc);
  MemoryContextSwitchTo(mcxt);
  initstree(S);
  /* find MUMs */
  if (uniqueinref)
    findmumcandidates(&S->tree, minmatchlength, processmum,
        rsi, query, querylen, 0);
  else
    findmaxmatches(&S->tree, minmatchlength, processmum,
        rsi, query, querylen, 0);
  PG_RETURN_NULL();
}


/* ======= maxmatchcount ======= */

static Sint processmumcount (void *info, Uint matchlength,
    Uint subjectstart, Uint seqnum, Uint querystart) {
  MUM_Count *mumcount = (MUM_Count *) info;
  (void) matchlength; (void) subjectstart;
  (void) seqnum; (void) querystart;
  mumcount->nmatches++;
  return mumcount->anymatch;
}

PG_FUNCTION_INFO_V1(maxmatchcount);
Datum maxmatchcount (PG_FUNCTION_ARGS) {
  MUM_Count mumcount;
  Stree *S = (Stree *) PG_GETARG_VARLENA_P(0);
  text *q = PG_GETARG_TEXT_P(1);
  Uchar *query = (Uchar *) VARDATA(q);
  Uint querylen = VARSIZE(q) - VARHDRSZ;
  bool uniqueinref = PG_GETARG_BOOL(2);
  Uint minmatchlength = (Uint) PG_GETARG_UINT32(3);
  mumcount.nmatches = 0;
  mumcount.anymatch = PG_GETARG_BOOL(4);
  if (!isseqvalid(query, querylen))
    elog(ERROR, "invalid symbol or empty sequence");
  if (minmatchlength == 0) minmatchlength = querylen;
  if (querylen < minmatchlength) /* early return? */
    PG_RETURN_INT32(0);
  initstree(S);
  if (uniqueinref)
    findmumcandidates(&S->tree, minmatchlength, processmumcount,
        &mumcount, query, querylen, 0);
  else
    findmaxmatches(&S->tree, minmatchlength, processmumcount,
        &mumcount, query, querylen, 0);
  PG_RETURN_INT32(mumcount.nmatches);
}

